from math import inf
from airpage.core._basic import *

# directly import will cause Exception:
class __:
    encoding = 'utf-8'
_stdout = sys.stdout
sys.stdout = __()

from py_trees.common import Status
from py_trees.behaviour import *
from py_trees.composites import *
from py_trees.decorators import *

sys.stdout = _stdout

''' ----------------------------------------- '''

_id_dict = {}

def GetNewId(cls_name:str):
    get = _id_dict.get(cls_name)
    if get is None:
        get = -1

    _id_dict[cls_name] = get + 1
    return get + 1

def GetNewName(cls_name:str):
    _id = GetNewId(cls_name)
    return f'{cls_name}{_id}'

''' -------------------------------- GetNewXXX End ------------------------------------- '''


class NoneException(Exception):
    ...

class ConditionNotComplete(Exception):
    ...

''' -------------------------------- Exception End ------------------------------------- '''

'''
行为树的三种返回值
返回Status非常严谨，诚然优秀。
但我更喜欢随意有效的返回值，因此，ap中用于行为树的函数的返回值必须为以下值。
* 返回None将抛出NoneException
* 返回非bool and 非... 将会将其解释为bool
** 由以上注意事项可知，在ap中返回Status中任何值都会被视作True。请注意这一点
'''
SUCCESS = True
RUNNING = ...
FAILURE = False

def ToStatus(ap_status, *, fn=None):

    if ap_status is None:
        txt_fn = f'(fn={fn.__name__})' if fn is not None else 'ToStatus'
        txt = f"\n\t{txt_fn} has returned None. which is not valid, please return bool or ..."
        raise NoneException(txt)

    if ap_status == ...:
        return Status.RUNNING

    else:
        ap_status = bool(ap_status)
        if ap_status:
            return Status.SUCCESS
        else:
            return Status.FAILURE

def FromStatus(status):
    if status == Status.INVALID:
        return None
    elif status == Status.SUCCESS:
        return True
    elif status == Status.FAILURE:
        return False
    elif status == Status.RUNNING:
        return ...
    else:
        raise TypeError(f"Unknown statustype: {status}")

''' -------------------------------- Status End ------------------------------------- '''
# redirect

class Action(Behaviour):
    """
    叶子节点Action。
    * 这个class的任务是包装任意函数, 使其可以塞到行为树中
    * 被包装的函数的返回值需要为 bool或是...
    """
    def setup(self, target:typing.Callable, *args: typing.Any, _sys_before_=None, _sys_after_=None, **kwargs: typing.Any) -> None:
        """
        指定需要action包装的fn
        :param target: 该action所包装的fn
        :param args: fn的参数
        :param kwargs: fn的kv参数
        :return:
        """
        self.target = target
        self.target_args = args
        self.target_kwargs = kwargs

        # AP Part
        self._before = _sys_before_
        self._after = _sys_after_

    def update(self) -> common.Status:
        if self._before:
            self._before()
        result = self.target(*self.target_args, **self.target_kwargs)
        if self._after:
            self._after()
        result = ToStatus(result, fn=self.target)
        # print(result)
        self.feedback_message = f'{self} return {result}'
        return result

    def __str__(self):
        args = f", args={self.target_args}" if len(self.target_args) else ''
        kwargs = f", kwargs={self.target_kwargs}" if len(self.target_kwargs) else ''
        return f"{self.__class__.__name__}:{self.name}(fn={self.target.__name__}{args}{kwargs})"

# class ActionSleep(Behaviour):
#     def setup(self, duration=1.0, **kwargs: typing.Any) -> None:
#         self.duration = duration
#
#     def initialise(self) -> None:
#         self.start_time = time.time()
#
#     def update(self) -> common.Status:
#         if time.time() >= (self.start_time + self.duration):
#             return Status.SUCCESS
#         return Status.RUNNING

class Condition(Action):
    """
    叶子节点Condition。
    * 这个class的任务是包装任意函数, 使其可以塞到行为树中
    * 被包装的函数的返回值需要为 bool
    ** Condition节点不能返回..., 否则抛出ConditionNotComplete
    """
    def update(self) -> common.Status:
        result = self.target(*self.target_args, **self.target_kwargs)
        if result == ...:
            txt = f"\n\t{self.target.__name__} has returned RUNNING in a Condition Node."
            raise ConditionNotComplete(txt)
        result = ToStatus(result, fn=self.target)
        self.feedback_message = f'{self} return {result}'
        return result


class ThreadParallel(Parallel):
    """
    这个组合节点是在composite.Parallel的基础上修改的
    Parallels enable a kind of spooky at-a-distance concurrency.

    .. graphviz:: dot/parallel.dot

    A parallel ticks every child every time the parallel is itself ticked.
    The parallelism however, is merely conceptual. The children have actually been
    sequentially ticked, but from both the tree and the parallel's purview, all
    children have been ticked at once.

    The parallelism too, is not true in the sense that it kicks off multiple threads
    or processes to do work. Some behaviours *may* kick off threads or processes
    in the background, or connect to existing threads/processes. The behaviour itself
    however, merely monitors these and is itself encosced in a py_tree which only ever
    ticks in a single-threaded operation.

    * Parallels will return :data:`~py_trees.common.Status.FAILURE` if any
      child returns :py:data:`~py_trees.common.Status.FAILURE`
    * Parallels with policy :class:`~py_trees.common.ParallelPolicy.SuccessOnAll`
      only returns :py:data:`~py_trees.common.Status.SUCCESS` if **all** children
      return :py:data:`~py_trees.common.Status.SUCCESS`
    * Parallels with policy :class:`~py_trees.common.ParallelPolicy.SuccessOnOne`
      return :py:data:`~py_trees.common.Status.SUCCESS` if **at least one** child
      returns :py:data:`~py_trees.common.Status.SUCCESS` and others are
      :py:data:`~py_trees.common.Status.RUNNING`
    * Parallels with policy :class:`~py_trees.common.ParallelPolicy.SuccessOnSelected`
      only returns :py:data:`~py_trees.common.Status.SUCCESS` if a **specified subset**
      of children return :py:data:`~py_trees.common.Status.SUCCESS`

    Policies :class:`~py_trees.common.ParallelPolicy.SuccessOnAll` and
    :class:`~py_trees.common.ParallelPolicy.SuccessOnSelected` may be configured to be
    *synchronised* in which case children that tick with
    :data:`~py_trees.common.Status.SUCCESS` will be skipped on subsequent ticks until
    the policy criteria is met, or one of the children returns
    status :data:`~py_trees.common.Status.FAILURE`.

    Parallels with policy :class:`~py_trees.common.ParallelPolicy.SuccessOnSelected` will
    check in both the :meth:`~py_trees.behaviour.Behaviour.setup` and
    :meth:`~py_trees.behaviour.Behaviour.tick` methods to to verify the
    selected set of children is actually a subset of the children of this parallel.

    .. seealso::
       * :ref:`Context Switching Demo <py-trees-demo-context-switching-program>`
    """

    def __init__(
        self,
        name: str,
        policy: common.ParallelPolicy.Base,
        thread_num: int = 2 * cpu_count,
        children: typing.Optional[typing.List[behaviour.Behaviour]] = None,
    ):
        """
        Initialise the behaviour with name, policy and a list of children.

        Args:
            name: the composite behaviour name
            policy: policy for deciding success or otherwise (default: SuccessOnAll)
            thread_num: max thread num, it's better to be decided by your cpu_count
            children: list of children to add
        """
        super(ThreadParallel, self).__init__(name, policy, children)
        self.thread_num = thread_num

    def initialise(self) -> None:
        super().initialise()
        self.pool = ThreadPoolExecutor(max_workers=self.thread_num)

    def tick(self) -> typing.Iterator[behaviour.Behaviour]:
        """
        Tick over the children.

        Yields:
            :class:`~py_trees.behaviour.Behaviour`: a reference to itself or one of its children

        Raises:
            RuntimeError: if the policy configuration was invalid
        """
        self.logger.debug("%s.tick()" % self.__class__.__name__)
        self.validate_policy_configuration()

        # reset
        if self.status != common.Status.RUNNING:
            self.logger.debug("%s.tick(): re-initialising" % self.__class__.__name__)
            for child in self.children:
                # reset the children, this ensures old SUCCESS/FAILURE status flags
                # don't break the synchronisation logic below
                if child.status != common.Status.INVALID:
                    child.stop(common.Status.INVALID)
            self.current_child = None
            # subclass (user) handling
            self.initialise()

        # nothing to do
        if not self.children:
            self.current_child = None
            self.stop(common.Status.SUCCESS)
            yield self
            return


        tasks = [self.pool.submit(lambda : list(child.tick())) for child in self.children]
        _, results = wait_results(tasks, timeout=inf)

        for item in results:
            for node in item:
                yield node

        # determine new status
        new_status = common.Status.RUNNING
        self.current_child = self.children[-1]
        try:
            failed_child = next(
                child
                for child in self.children
                if child.status == common.Status.FAILURE
            )
            self.current_child = failed_child
            new_status = common.Status.FAILURE
        except StopIteration:
            if type(self.policy) is common.ParallelPolicy.SuccessOnAll:
                if all([c.status == common.Status.SUCCESS for c in self.children]):
                    new_status = common.Status.SUCCESS
                    self.current_child = self.children[-1]
            elif type(self.policy) is common.ParallelPolicy.SuccessOnOne:
                successful = [
                    child
                    for child in self.children
                    if child.status == common.Status.SUCCESS
                ]
                if successful:
                    new_status = common.Status.SUCCESS
                    self.current_child = successful[-1]
            elif type(self.policy) is common.ParallelPolicy.SuccessOnSelected:
                if all(
                    [c.status == common.Status.SUCCESS for c in self.policy.children]
                ):
                    new_status = common.Status.SUCCESS
                    self.current_child = self.policy.children[-1]
            else:
                raise RuntimeError(
                    "this parallel has been configured with an unrecognised policy [{}]".format(
                        type(self.policy)
                    )
                )
        # this parallel may have children that are still running
        # so if the parallel itself has reached a final status, then
        # these running children need to be terminated so they don't dangle
        if new_status != common.Status.RUNNING:
            self.stop(new_status)
        self.status = new_status
        yield self

    def stop(self, new_status: common.Status = common.Status.INVALID) -> None:
        """
        Ensure that any running children are stopped.

        Args:
            new_status : the composite is transitioning to this new status
        """
        # close pool
        if self.pool:
            self.pool.shutdown()

        super(ThreadParallel, self).stop(new_status)



class Succeedor(Decorator):
    def update(self) -> common.Status:
        return Status.SUCCESS

class Repeater(Decorator):
    """
    这个类基于Repeat
    Repeat.

    :data:`~py_trees.common.Status.SUCCESS` is
    :data:`~py_trees.common.Status.RUNNING` up to a specified number at
    which point this decorator returns :data:`~py_trees.common.Status.SUCCESS`.

    :data:`~py_trees.common.Status.FAILURE` is always
    :data:`~py_trees.common.Status.FAILURE`.

    Args:
        child: the child behaviour or subtree
        times: repeat max times (-1 to repeat indefinitely)
        until: repeat until inner_node return True/False.
    """

    def __init__(self, name: str, child: behaviour.Behaviour, times: int=-1, until:bool=None):
        super().__init__(name=name, child=child)
        self.left_times = 0
        self.max_times = inf if times <= 0 else times
        self.until = None if until is None else ToStatus(until)

    def initialise(self) -> None:
        """Reset the currently registered number of successes."""
        self.left_times = self.max_times

    def tick(self) -> typing.Iterator[behaviour.Behaviour]:
        """
        Manage the decorated child through the tick.

        Yields:
            a reference to itself or one of its children
        """
        self.logger.debug("%s.tick()" % self.__class__.__name__)
        # initialise just like other behaviours/composites
        if self.status != common.Status.RUNNING:
            self.initialise()
        # interrupt proceedings and process the child node
        # (including any children it may have as well)
        for node in self.decorated.tick():
            yield node
        # resume normal proceedings for a Behaviour's tick
        new_status = self.update()
        if new_status not in list(common.Status):
            self.logger.error(
                "A behaviour returned an invalid status, setting to INVALID [%s][%s]"
                % (new_status, self.name)
            )
            new_status = common.Status.INVALID
        if new_status != common.Status.RUNNING:
            self.stop(new_status)
        self.status = new_status
        yield self

    def update(self) -> common.Status:
        """
        Repeat until the nth consecutive success.

        Returns:
            :data:`~py_trees.common.Status.SUCCESS` on nth success,
            :data:`~py_trees.common.Status.RUNNING` on running, or pre-nth success
            :data:`~py_trees.common.Status.FAILURE` failure.
        """

        if self.decorated.status == Status.RUNNING:
            return Status.RUNNING
        else:
            self.left_times -= 1

            if self.until is not None and self.decorated.status == self.until:
                self.feedback_message = f"success end until:{self.until}"
                return Status.SUCCESS
            elif self.until is not None and self.left_times <= 0:
                self.feedback_message = f"failure end for ran out of times and not meet until:{self.until}"
                return Status.FAILURE
            elif self.left_times <= 0:
                self.feedback_message = f"success end for ran out of times"
                return Status.SUCCESS
            else:
                return Status.RUNNING


class Reflictor(Decorator):
    def setup(self, status_x, status_y, **kwargs: typing.Any) -> None:
        self.status_x = status_x
        self.status_y = status_y

    def update(self) -> common.Status:
        """
        Reflect :data:`~py_trees.common.Status.RUNNING` as :data:`~py_trees.common.Status.SUCCESS`.

        Returns:
            the behaviour's new status :class:`~py_trees.common.Status`
        """
        if self.decorated.status == self.status_x:
            self.feedback_message = f"running is {self.status_x}" + (
                " [%s]" % self.decorated.feedback_message + f" -- reflict to {self.status_y}"
                if self.decorated.feedback_message
                else ""
            )
            return self.status_y
        self.feedback_message = self.decorated.feedback_message
        return self.decorated.status

''' -------------------------------- Class:Leaf End ------------------------------------- '''
Behaviour.__call__ = Behaviour.tick_once

def get_gradparent_calling_function(_reback_level=2):
    """finds the calling function's calling function in many decent cases."""
    fr = sys._getframe(_reback_level)   # inspect.stack()[2][0]
    co = fr.f_code
    for get in (
        lambda:fr.f_globals[co.co_name],
        lambda:getattr(fr.f_locals['self'], co.co_name),
        lambda:getattr(fr.f_locals['cls'], co.co_name),
        lambda:fr.f_back.f_locals[co.co_name], # nested
        lambda:fr.f_back.f_locals['func'],  # decorators
        lambda:fr.f_back.f_locals['meth'],
        lambda:fr.f_back.f_locals['f'],
        ):
        try:
            func = get()
        except (KeyError, AttributeError):
            pass
        else:
            if func.__code__ == co:
                return func
    raise AttributeError("func not found")

def action(fn, *args, name=None, pack=False, _reback=2, **kwargs):
    """
    基于Behaviour的Action，运行时实际执行一个fn对象。该fn需要返回SUCCESS FAILURE RUNNING中的一种值作为该节点的状态
    :param fn: 实际需要执行的fn对象
    :param args: fn对象需要的args参数
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param kwargs: 对应的Behaviour节点初始化时可选的额外kv参数
    :return: Behaviour
    """
    act = Action(name if name else fn.__name__)
    act.setup(fn, *args, **kwargs)
    if pack:
        act.packer = get_gradparent_calling_function(_reback).__name__
    return act

def condition(fn, *args, name=None, pack=False, _reback=2, **kwargs):
    """
    基于Action，两者唯一的区别是condition不能返回RUNNING
    :param fn: 实际需要执行的fn对象
    :param args: fn对象需要的args参数
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param kwargs: 对应的Behaviour节点初始化时可选的额外kv参数
    :return: Behaviour
    """
    cond = Condition(name if name else fn.__name__)
    cond.setup(fn, *args, **kwargs)
    if pack:
        cond.packer = get_gradparent_calling_function(_reback).__name__
    return cond

# --- Leaf ---

def fallback(*nodes, name=None, pack=False, **kwargs):
    """
    依次执行子节点组, 直到某个子节点返回True/RUNNING时返回True/RUNNING。如果所有子节点都返回了False，那么返回False
    :param *nodes: 子节点组
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param kwargs: 对应的组合节点初始化时可选的额外kv参数
    :return: Composite
    """
    new_node = Selector(name if name else GetNewName('Fallback'), memory=True, children=nodes)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def sequence(*nodes, name=None, pack=False, **kwargs):
    """
    依次执行子节点组, 直到某个子节点返回False/RUNNING时返回False/RUNNING。如果所有子节点都返回了True，那么返回True
    :param *nodes: 子节点组
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param kwargs: 对应的组合节点初始化时可选的额外kv参数
    :return: Composite
    """
    new_node = Sequence(name if name else GetNewName('Sequence'), memory=True, children=nodes)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def parallel(*nodes, name=None, pack=False, policy='and', thread_num=cpu_count+1, **kwargs):
    """
    并行运行所有子节点，等待所有子节点结束。返回结果与policy有关.
    :param *nodes: 子节点组
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param policy: 'and' 'or'
        * 'and' 所有子节点的结果都为SUCCESS，才返回SUCCESS，否则返回FAILURE
        ** and模式下如果出现FAILURE会直接无视RUNNING返回FAILURE
        * 'or' 任意子节点的结果为SUCCESS，就返回SUCCESS，除非所有子节点都返回FAILURE才返回FAILURE
        ** or模式下如果出现SUCCESS会直接无视RUNNING返回SUCCESS
        *** 否则返回RUNNING
    :param thread_num: 最大线程数，默认cpu_count+1
    :param kwargs: 对应的组合节点初始化时可选的额外kv参数
    :return: Composite
    """
    if policy == 'and':
        policy = common.ParallelPolicy.SuccessOnAll()
    elif policy == 'or':
        policy = common.ParallelPolicy.SuccessOnOne()
    else:
        raise TypeError(f"Unknown typekind of policy: {policy}")

    new_node = ThreadParallel(name if name else GetNewName('Parallel'), policy, thread_num, nodes)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

# --- Composite ---

def succeedor(*nodes, name=None, pack=False, **kwargs):
    """
    不论子节点返回了什么，总是像上一节点返回True
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    new_node = Succeedor(name if name else GetNewName('Succeedor'), node)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def repeater(*nodes, name=None, pack=False, times=-1, until=None, **kwargs):
    """
    重复循环子节点n次后返回True。如果定义了until, 那么在耗尽循环次数n前满足until时返回True, 否则返回False。子节点返回RUNNING时也返回RUNNING
    * 只有返回非RUNNING时才视作完成了一次循环
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param times: 最大循环次数，传入-1视作无限循环
    :param until: 循环跳出条件, 当子节点的返回值与该条件相符时会提前结束循环。默认None
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    new_node = Repeater(name if name else GetNewName('Repeater'), node, times, until)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def inverter(*nodes, name=None, pack=False, **kwargs):
    """
    对子节点的结果进行取反。子节点返回RUNNING时也返回RUNNING
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    new_node = Inverter(name if name else GetNewName('Inverter'), node)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def validator(condition_fn:typing.Callable, *nodes,name=None, pack=False, blackboard_keys:typing.List[str]=None, **kwargs):
    """
    在执行子节点前先验证条件函数的返回值. 如果条件函数返回True，则返回子节点的结果，否则返回False
    :param condition_fn: 验证函数，在正式执行子节点前，先运行验证函数。若bool(验证函数值返回值)为True, 才会执行子节点
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param blackboard_keys: provide read access for the conditional function to these keys
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    new_node = EternalGuard(name if name else GetNewName('Validator'), node, condition, blackboard_keys)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def timeouter(*nodes, name=None, pack=False, duration=5.0, **kwargs):
    """
    带时限执行子节点。如果超时将尝试提前结束子节点并返回False，否则返回子节点的返回值
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param duration: 时限值(单位s)，超过此值后悔提前结束子节点运行
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    new_node = Timeout(name if name else GetNewName('Timeouter'), node, duration)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def counter(*nodes, name=None, pack=False, **kwargs):
    """
    记录子节点各种状态出现的次数, 无法作为root节点。直接调用此对象将返回一段报告文本以显示子节点运行的情况
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    new_node = Count(name if name else GetNewName('Counter'), node, duration)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def oneshoter(*nodes, name=None, pack=False, policy='pass', **kwargs):
    """
    一直执行子节点，直到子节点返回非RUNNING的情况才算结束。返回值取决于policy参数的值
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param policy: 'pass' 'any'
        * 'pass' 返回子节点的返回值(比如子节点返回SUCCESS，那么oneshot也返回SUCCESS)
        * 'any' 不关心子节点执行的结果，只要执行完毕，就返回SUCCESS
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if policy == 'any':
        policy = common.OneShotPolicy.ON_COMPLETION
    elif policy == 'pass':
        policy = common.OneShotPolicy.ON_SUCCESSFUL_COMPLETION
    else:
        raise TypeError(f"Unknown typekind of policy: {policy}")

    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    new_node = OneShot(name if name else GetNewName('OneShoter'), node, policy)
    new_node.setup(**kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return new_node

def reflictor(*nodes, name=None, pack=False, x=RUNNING, y=FAILURE, **kwargs):
    """
    将子节点的某种返回值映射为另一返回值(x->y)
    :param *nodes: 子节点。
            * 当子节点个数超过两个时自动为其创建sequence
    :param name: 当前节点的名称，传入None时使用默认名称(注:名称和在渲染图中的名称不是一个值)
    :param pack: 是否打包该节点及其子节点为一个模块
            * 主要在渲染行为树时生效
            * 被打包的行为树节点会被渲染器视作一个子节点
    :param x: 输入的目标返回值。如果子节点返回值属于x的类型，那么将返回y
    :param y: 转换后的返回值
    :param kwargs: 对应的装饰节点初始化时可选的额外kv参数
    :return: Decorator
    """
    if len(nodes) > 1:
        node = sequence(*nodes)
    else:
        node = nodes[0]
    rf = Reflictor(name if name else GetNewName('Reflictor'), node)
    rf.setup(ToStatus(x), ToStatus(y), **kwargs)
    if pack:
        new_node.packer = get_gradparent_calling_function().__name__
    return rf

# --- Decorator End ---
def SetTreeEntryence(entryence_fn):
    Behaviour.__call__ = entryence_fn

# --- Run ---

def SetTreeRender(render_fn):
    Behaviour.render = render_fn

''' -------------------------------- function API END ------------------------------------- '''

if __name__ == '__main__':
    def Test_Polling(btree: Behaviour, oneshot=False):
        """
        轮询行为树, 并返回轮询结果
        * 具有AP特色
        :param btree: BehaviourTree
        :return: bool or ...
        """
        while True:
            btree.tick_once()
            result = btree.status

            if result != Status.RUNNING or not oneshot:
                break

        return FromStatus(result)

    SetTreeEntryence(Test_Polling)

    _ = [False]
    def Test(_id, slp_time=0.5):
        print(f"Enter fn-Test{_id}")
        time.sleep(slp_time)
        print(f'after sleep {slp_time}s, fn-Test{_id} finished.')
        _[0] = not _[0]
        return _[0]

    node = repeater(
        action(Test, 0),
        times=3,
        until=False
    )

    print("--- start ---")

    for i in range(1):
        node(oneshot=True)
        print(node.status)
        time.sleep(0.5)
    print('\n--- finish ---')


import numpy as np
from cnocr import CnOcr
from airpage.core._basic import *
from airtest.aircv import aircv
import warnings
import tempfile

warnings.filterwarnings("ignore")

class APOcr:
    def __init__(self):
        self.ocr = CnOcr()

    @staticmethod
    def CountIOU(RecA, RecB):
        """
        计算IOU(重叠面积大小问题)
        :param RecA: (x0, y0, x1, y1)
        :param RecB: (x0, y0, x1, y1)
        :return: float 0 - 1
        """
        xA = max(RecA[0], RecB[0])
        yA = max(RecA[1], RecB[1])
        xB = min(RecA[2], RecB[2])
        yB = min(RecA[3], RecB[3])
        # 计算交集部分面积
        interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)
        # 计算预测值和真实值的面积
        RecA_Area = (RecA[2] - RecA[0] + 1) * (RecA[3] - RecA[1] + 1)
        RecB_Area = (RecB[2] - RecB[0] + 1) * (RecB[3] - RecB[1] + 1)
        # 计算IOU
        iou = interArea / float(RecA_Area + RecB_Area - interArea)
        return iou

    def NMS(self, dets, thresh=0.5):
        """
        非极大抑制
        :param dets: nd.array [[x0, y0, x1, y1, score], ...]
        :param tr: float
        :return:
        """
        x1 = dets[:, 0]
        y1 = dets[:, 1]
        x2 = dets[:, 2]
        y2 = dets[:, 3]
        scores = dets[:, 4]

        # 每一个检测框的面积
        areas = (x2 - x1 + 1) * (y2 - y1 + 1)

        # 按照score置信度降序排序
        order = scores.argsort()[::-1]

        keep = []  # 保留的结果框集合

        while order.size > 0:
            i = order[0]
            keep.append(i)  # 保留该类剩余box中得分最高的一个
            # 得到相交区域,左上及右下
            xx1 = np.maximum(x1[i], x1[order[1:]])
            yy1 = np.maximum(y1[i], y1[order[1:]])
            xx2 = np.minimum(x2[i], x2[order[1:]])
            yy2 = np.minimum(y2[i], y2[order[1:]])
            # 计算相交的面积,不重叠时面积为0
            w = np.maximum(0.0, xx2 - xx1 + 1)
            h = np.maximum(0.0, yy2 - yy1 + 1)
            inter = w * h
            # 计算IoU：重叠面积 /（面积1+面积2-重叠面积）
            ovr = inter / (areas[i] + areas[order[1:]] - inter)
            # 保留IoU小于阈值的box
            inds = np.where(ovr <= thresh)[0]
            order = order[inds + 1]  # 因为ovr数组的长度比order数组少一个,所以这里要将所有下标后移一位
        return keep

    def ap_ocr(self, tpl:Template):
        """
        寻找一个tpl对应的屏幕区域存在的文字
        按照逐行逐行的顺序返回列表
        :param tpl: Template
        :return: list of text
        """
        fp = tpl.filepath
        area = ap_area(tpl, toint=True)
        screen = GetScreen()
        tmp_fname = GetTempName() + '.png'
        try:
            roi = screen[area[1]:area[3], area[0]:area[2]]
        except:
            error("RoiError", f"Area:{area} does not match with Screen_HW:{screen.shape[:2]}")

        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_fpath = os.path.join(tmpdirname, tmp_fname)

            aircv.imwrite(r'C:\Users\Administrator\Desktop\test\tmp.png', roi)

            aircv.imwrite(tmp_fpath, roi)
            result = self.ocr.ocr(fp)

        values, score_areas = [], []

        for each in result:
            if not isinstance(each, dict):
                continue
            value = each.get('text')
            score = each.get('score')
            position = each.get('position')

            if value is not None and score is not None and position is not None:
                area = *position[0], *position[2]

                values.append(value)
                score_areas.append((*area, score))

        if not len(values):
            return []

        # 非极大抑制
        _lsa = self.NMS(np.array(score_areas))

        # 还原顺序
        order = []
        for each in _lsa:
            for i, item in enumerate(score_areas):
                _flag = False
                for t in range(item):
                    if item[t] != each[t]:
                        _flag = True
                        break
                if not _flag:
                    order.append(i)
                    break

        order.sort(reverse=False)  # 升序排序

        return [values[i] for i in order]











_ocr = [None]
def GetOcr():
    if _ocr[0] is None:
        _ocr[0] = APOcr()
    return _ocr[0]










import os.path

from airpage.core import *

new_main_content = f"""__author__ = '{os.getlogin()}'
from airpage import GetEnviron
env = GetEnviron(__file__)
''' ------------------ Environment Ready -------------------- '''
# Create your script here:
from page import *


"""
new_page_content = f"""from backend import *
Page('shut', Not(IDENTITY))     # page when game not start.
Page('main', IDNETITY)          # main page (like menu).

''' ------------------ Page Definition End -------------------- '''
# @transfer(page_a, page_b)
# Page.LinkWith(page_a, page_b, fn_a2b, fn_b2a)

Page.LinkWith(
    'shut',
    'main',
    sequence(XXX),  # How to start game
    sequence(XXX),  # How to close game
)

''' ------------------ Page Transition End -------------------- '''
# @page(page_name)
# Page.LoginTask(page_name, fn_task)

"""
new_backend_content = f"""from airpage import *
# define custom functions


"""
def NewProject(project_path:str):
    """
    新建airpage工程。
    :param project_path: 工程路径，可以是空的，也可以是已经刚刚由airtest创建的工程
    :return: bool 成功返回True，失败返回False
    """
    if project_path[-4:].lower() != '.air':
        error("TypeError", "Project path must be endwith '.air'")

    if os.path.exists(project_path):
        warn(f"Exists {project_path}, Failed to create.")
        return False
    else:
        ''' 不存在项目，可以开始快乐创建空项目了 '''
        os.makedirs(project_path)
        assert os.path.exists(project_path), f"Failed to create directory:{project_path}"

        # 创建主脚本
        fname = os.path.basename(project_path)
        if fname[-4:].lower() == '.air':
            fname = fname[:-4]
        with open(os.path.join(project_path, fname + '.py'), 'w') as f:
            f.write(new_main_content)

        # 创建page脚本
        with open(os.path.join(project_path, 'page.py'), 'w') as f:
            f.write(new_page_content)
        return True

def CMD_NewProject(*argv):
    if len(argv):
        return NewProject(*argv)
    else:
        print("Missing Parameter.\n", help(NewProject))

def CMD_Version(*argv):
    import pkg_resources
    print(f"airpage version:{pkg_resources.get_distribution('airpage').version}")

if __name__ == '__main__':
    CMD_Version()


from airpage.core._basic import *
from airpage.core._btree import SUCCESS, FAILURE, RUNNING, FromStatus, ToStatus
from airpage.core._btree import action, condition, Behaviour
from airpage.core._btree import sequence, fallback, parallel
from airpage.core._btree import succeedor, repeater, inverter, validator, oneshoter, timeouter, counter, reflictor

# from airpage._ocr import APOcr, GetOcr

class Not:
    """
    定义了一种not容器
    * 只能存储一个元素
    * 用在Identity中，用于表示Not逻辑
    """
    def __init__(self, item):
        self.item = item

    def __bool__(self):
        return not bool(item)

    def __len__(self):
        return 1

    def __iter__(self):
        return iter([self.item])


class Identity:
    """
    定义了一个基于tpl的逻辑表达
    * 垄断了所有的检测

    简单来说，一个Identity实例是一个and or not的逻辑表达式，只不过表达元素是tpl对象
    ---------------------------------------------
    :param identity: list|tuple|Not|tpl airpage定义的逻辑组合，可以映射成or and not
            list -> or
            tuple -> and
            Not -> not
            tpl -> single
            @ example:
                not( (a and b) or (c and d) )
                可以通过list|tuple|set这一映射转变为
                Not( [(a, b), (c, d)] )
    """
    def __init__(self, identity):
        self.express, self.identity = self._build_express(identity)
        self.identity_keys = list(self.identity.keys())
        self.identity_values = list(self.identity.values())

    @staticmethod
    def _build_express(identity):
        """
        这个函数根据identity生成对应的express和对应的数据对象
        :param identity: list|tuple|set airpage定义的逻辑组合，可以映射成or and not
        :return: (express, identity:dict)
        """
        # 计数器, 可以通过这个函数获取不重复的索引
        _cnt = [0]
        def _cnt_():
            _cnt[0] += 1
            return _cnt[0]

        _fdict = {}
        def _inner_(item_node):
            """
            :param item_node:
            :return: [_id, ...]
            """
            _id, _exp = 0, ''
            if isinstance(item_node, bool):  # 函数出口
                return str(item_node)
            elif isinstance(item_node, list):
                _exp = 'or'
            elif isinstance(item_node, tuple):
                _exp = 'and'
            elif isinstance(item_node, Not):
                _exp = 'not'
            elif isinstance(item_node, Template):  # 函数出口
                _id = f"__{_cnt_()}"
                _fdict[_id] = item_node
                return _id
            else:
                error("UnknownType", f"Expected bool|List|Tuple|Not|Template, but get {type(item_node)}")

            tmp, fmt = '', f" {_exp} "
            len_fmt = len(fmt)
            for each in item_node:
                tmp += fmt + _inner_(each)
            if tmp and _exp != 'not':
                tmp = tmp[len_fmt:]

            return f"({tmp})"

        return _inner_(identity), _fdict

    def check(self, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True):
        """
        检测当前画面是否满足Identity的要求
        :return: bool
        """
        _exp = self.express

        values = ap_wait(self.identity_values, timeout=timeout, interval=interval, threshold=threshold, snapshot=snapshot)

        # --- fill expression ---

        for k ,v in zip(self.identity_keys, values):
            _exp = _exp.replace(k, str(bool(v)))

        # --- eval expression ---
        try:
            return eval(_exp)
        except:
            warn(f"Failed check: {k} (Happen in: eval({_exp}))")
            return False

    def burstCheck(self, timeout=ST.FIND_TIMEOUT, interval=0.5, snapshot=False):
        """
        爆发式检测
        * 在短时间内没有任何!'操作'!
        * 在短时间内需要大量检测
        * 需要外部库内调用check方法(传参困难)
        :return:
        """
        return self.check(timeout=timeout, interval=interval, snapshot=snapshot)

    @classmethod
    def Check(cls):
        if isinstance(identity, Identity):
            return identity.check(timeout=timeout, interval=interval, threshold=threshold, snapshot=snapshot)
        else:
            return Identity(identity).check(timeout=timeout, interval=interval, threshold=threshold, snapshot=snapshot)

    def __bool__(self):
        return self.check()


_current_environment = [None]
_flag_environment = [False]

class APEnviron:
    """
    airpage使用的全局坏境, 启动程序时会自动创建一个默认的全局环境。一个全局环境默认只负责一款程序的自动化。
    """
    def __init__(self, project_path):
        self._pages = {}  # Page
        self._ppath = project_path

        # ...
        self.LoadProject(project_path)

    def LoadProject(self, main_file_path):
        auto_setup(main_file_path)
        sys.path.append(os.path.dirname(main_file_path))
        self._ppath = os.path.dirname(main_file_path)

    @property
    def project_path(self):
        return self._ppath

    def __del__(self):
        self.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type:typing.Type, exc_val, exc_tb):
        self.close()
        if exc_type is not None:
            raise exc_type(exc_val, exc_tb)
        return True

    def open(self, timeout:int=0xf) -> None:
        """
        将self设为当前环境, 并阻塞其他环境调用open方法
        :param timeout: 阻塞的最大等待时间，超时后抛出异常
        :return: None
        """
        if _current_environment[0] is self:
            return

        # wait
        timeout_point = time.time() + timeout
        while _flag_environment[0]:
            time.sleep(ST.OPDELAY)
            if time.time() > timeout_point:
                raise error("TimeoutError", "ran out of time for waiting APEnviron.open")

        # lock
        _flag_environment[0] = True
        _current_environment[0] = self

    def close(self):
        """
        关闭自身环境
        :return: None
        """
        if _current_environment is not self:
            return

        # unlock
        _flag_environment[0] = False

        # reset
        _current_environment[0] = None

    ''' -------------------------------------------------------------------- '''

    def GetPage(self, page_name:str):
        """
        获取指定名称的Page对象, 如果不存在，返回None
        :param page_name: str特定page的名称
        :return: Page | None
        """
        return self._pages.get(page_name)

    def SetPage(self, page_name:str, page:typing.Any):
        """
        设置Page，可以覆盖，没有检查，使用时谨慎
        :param page_name: str特定page的名称
        :param page: Page对象
        :return: None
        """
        self._pages[page_name] = page


def GetEnviron(project_path=None) -> APEnviron:
    if _current_environment[0] is None:
        APEnviron(project_path).open()
    return _current_environment[0]

if __name__ == '__main__':
    def debug_test(a, b):
        print(a, b)
        return (a, b)


    act = action_debug_test(1, 2)

    AsyncRun(act)

    # print(test.target)


















import os
import sys
import time
import uuid
import files3
import random
import typing
import winreg
import platform

import pyautogui
from airtest.core.api import *
from airtest.core.cv import Predictor
from concurrent.futures import ThreadPoolExecutor

''' -------------------------------------------------- '''

cpu_count = os.cpu_count()
pyautogui.FAILSAFE=False


if platform.system() == 'Windows':
    # redirect user-environ to system-environ
    reg_path = r'SYSTEM\CurrentControlSet\Control\Session Manager\Environment'
    reg_key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_path)
    system_environment_variables = winreg.QueryValueEx(reg_key, 'Path')[0]
    os.environ["PATH"] = system_environment_variables

def GetTempName():
    """
    获取一个临时名称，所有临时名称之间不会重复
    :return: str
    """
    return str(uuid.uuid1())

class __Null__():
    def __bool__(self):
        return False
Null = __Null__()

class DeviceExceoption(Exception):
    ...

def GetDevice():
    if G.DEVICE is None:
        raise DeviceExceoption("\n\tDEVICE is NoneType. Please check your device connection and program entrance.")
    return G.DEVICE

def time2str(_time, _fmt_level=0) -> str:
    """
    将标准时间戳转化为项目中标准的字符串格式
    :Param _fmt_level:
        0  %Y-%m-%d
        1  %H:%M:%S
        2  %Y-%m-%d %H:%M:%S
    """
    '''Select'''
    if _fmt_level == 0:
        _fmt = "%Y-%m-%d"
    elif _fmt_level == 1:
        _fmt = "%H:%M:%S"
    elif _fmt_level == 2:
        _fmt = "%Y-%m-%d %H:%M:%S"
    else:
        raise Exception("[Basic - ERROR]: Unknown time-fmt type.")

    '''Translation'''
    time_struct = time.localtime(_time)
    date_string = time.strftime(_fmt, time_struct)
    return date_string

''' ------------------------------- Basic Functions End ---------------------------------------- '''

def label(desc):
    print(f'[Label]: {desc}')
    return True

def debug(*why):
    txt = '[Debug]:\n'
    for each in why:
        txt += '\t' + each + '\n'
    print(txt)
    # G.LOGGING.debug(txt)
    return True

def info(*why):
    txt = '[Info]:\n'
    for each in why:
        txt += '\t' + each + '\n'
    print(txt)
    # G.LOGGING.info(txt)
    return True

def warn(*why):
    txt = '[Warning]:\n'
    for each in why:
        txt += '\t' + each + '\n'
    print(txt)
    # G.LOGGING.warning(txt)
    return True

def error(err_type: str, *why):
    txt = "\n\n[{}]:".format(err_type)
    for each in why:
        txt += "\n\t" + str(each)
    txt += "\n"
    print(txt)
    # G.LOGGING.error(txt)
    raise type(err_type, (Exception, ), {})(txt)

''' ---------------------------------- Basic Information functions End ----------------------------------- '''
def ap_pos(record_pos_x, record_pos_y):
    """
    将airtest的相对百分比坐标转换为device的相对坐标
    airtest百分比坐标原点位于device的屏幕中心
    :param record_pos_x: 百分比坐标的x部分 -0.5 - 0.5
    :param record_pos_y: 百分比坐标的y部分 -0.5 - 0.5
    :return: tuple(x, y)
    """
    return Predictor.get_predict_point((record_pos_x, record_pos_y), GetDevice().get_current_resolution())

def ap_area(tpl:Template, toint=False):
    """
    根据template在当前的screen中确定对应的区域
    :param tpl: 根据template在当前的screen中确定对应的区域
    :return: tuple(x0, y0, x1, y1)
    """
    image_hw = tpl._imread().shape[:2]
    area = Predictor.get_predict_area(tpl.record_pos, (image_hw[1], image_hw[0]), tpl.resolution, GetDevice().get_current_resolution())
    if toint:
        area = int(area[0] + 0.5), int(area[1] + 0.5), int(area[2] + 0.5), int(area[3] + 0.5)
    return area

def wait_results(tasks, timeout=ST.FIND_TIMEOUT, interval=0.5):
    """
    等待由ThreadPoolExecutor创建的所有task完全执行结束
    :param tasks: list of (由ThreadPoolExecutor创建的所有task)
    :param timeout: 超时时间
    :param interval: 间隔时间
    :return: whether all success(all ok will return True, False else), list of result(No result will replace with Null)
    """
    flag, length = False, len(tasks)
    results = [Null for i in range(length)]

    timeout_time = time.time() + timeout

    while time.time() <= timeout_time:
        flag = True
        time.sleep(interval)
        for i, task in enumerate(tasks):
            if results[i] is Null:
                if task.done():
                    results[i] = task.result()
                else:
                    flag = False

        if flag:
            break


    return flag, results

_global_screens = {
    False: None  # 默认截图
}

def GlobalScreen_Update(sid=True):
    """
    更新屏幕截图
    :param sid: 更新截图到指定的screen index下(默认保存到False下)
    :return:
    """
    _global_screens[sid] = GetDevice().snapshot(filename=None, quality=ST.SNAPSHOT_QUALITY)
    return True

def GlobalScreen_Flush(sid=False):
    _global_screens[sid] = None
    return True

def GlobalScreen_Write(sid=False, _screen=None):
    _global_screens[sid] = _screen
    return True

def GlobalScreen_Read(sid=False):
    return _global_screens[sid]

def GetScreen(snapshot=True):
    tmp = _global_screens.get(snapshot)
    if snapshot is True or tmp is None:
        return GetDevice().snapshot(filename=None, quality=ST.SNAPSHOT_QUALITY)
    return tmp


def ap_wait(querys, *, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True):
    """
    升级版loop_find
    这段代码修改自airtest.core.cv.loop_find,
        * 去除其中的报错部分，改为返回None
        * 线程池调度，支持并发检测


    Args:
        fn_return: 返回函数，用于输出返回值
        querys: image template to be found in screenshot. Can be single tpl or multiple tpls
        timeout: time interval how long to look for the image template
        threshold: default is None
        interval: sleep interval before next attempt to find the image template
        snapshot: bool 是否进行snapshot。如果为False，那么优先使用globalscreen；但是如果globalscreen为None，那么还是会进行snapshot

    Returns:
        None or Position( or list of [None or Position])

    """

    isi_tpl = isinstance(querys, Template)
    isi_lt = isinstance(querys, (list, tuple))
    length = 1 if isi_tpl else len(querys)

    assert isi_tpl or (isi_lt and length > 0), f"Unexpected querys: {querys}"

    # ------

    G.LOGGING.info("Try finding: %s", querys)
    if length == 1:
        tpl = querys if isi_tpl else querys[0]

        start_time = time.time()
        while True:
            screen = GetScreen(snapshot)

            if screen is None:
                G.LOGGING.warning("Screen is None, may be locked")
            else:
                if threshold:
                    tpl.threshold = threshold
                match_pos = tpl.match_in(screen)
                if match_pos:
                    try_log_screen(screen)
                    return match_pos if isi_tpl else [match_pos]

            # 超时返回None，否则等待下次操作，
            if (time.time() - start_time) > timeout:
                try_log_screen(screen)
                G.LOGGING.info("Failed finding: %s", tpl)
                return False if isi_tpl else [False]
            else:
                time.sleep(interval)
    else:
        '''
        内存池的大小选取原则:
        一般来说，我们线程池究竟设置多大是基于要线程池任务来分析的，不同的任务类型，我们设置的方式当然也需要改变。
        通常的任务类型一般是：CPU密集型、IO密集型，对不同类型的任务我们就需要分配不同大小的线程池
        1)、CPU密集型
        这种任务我们要尽量使用较小的线程池，一般是Cpu核心数+1
        因为CPU密集型任务CPU的使用率很高，若开过多的线程，只能增加线程上下文的切换次数，带来额外的开销
        2)、IO密集型
        方法一：可以使用较大的线程池，一般CPU核心数 * 2
        IO密集型CPU使用率不高，可以让CPU等待IO的时候处理别的任务，充分利用cpu时间
        方法二：线程等待时间所占比例越高，需要越多线程。线程CPU时间所占比例越高，需要越少线程。
        '''
        pool = ThreadPoolExecutor(max_workers=min(cpu_count + 1, length))

        left = length
        results = [Null for i in range(length)]

        start_time = time.time()
        while left > 0:
            screen = GetScreen(snapshot)

            if screen is None:
                G.LOGGING.warning("Screen is None, may be locked")
            else:
                # 创建仍需match的任务表和对应映射表
                tasks, fntable = [ ], [ ]
                for i, tpl in enumerate(querys):
                    if results[i] is Null:
                        tmp = pool.submit(tpl.match_in, screen)
                        tasks.append(tmp)
                        fntable.append(i)

                _, task_results = wait_results(tasks, timeout, interval)

                for t, item in enumerate(task_results):
                    if item is not Null:
                        results[fntable[t]] = item
                        left -= 1

                # 超时对应元素返回None，否则等待下次操作，
                if (time.time() - start_time) > timeout:
                    try_log_screen(screen)

                    querys_timeout = []
                    for i, result in enumerate(results):
                        if result is Null:
                            results[i] = None  # 修改Null为None
                            querys_timeout.append(querys[i])

                    G.LOGGING.info("Failed finding: %s", querys_timeout)
                    break
                else:
                    time.sleep(interval)

        pool.shutdown()

        return results


def ap_exists(querys, *, threshold=None, snapshot=False):
    """
    like ap_wait, but timeout is 0, which mean this function only check screen once
    :param querys:
    :param threshold:
    :param snapshot:
    :return:
    """
    return ap_wait(querys, timeout=0, threshold=threshold, snapshot=snapshot)


def ap_touch(v, *, times=1, match=True, delay=None, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True, **kwargs):
    """
    这段代码修改自airtest.touch

    Perform the touch action on the device screen

    :param v: target to touch, either a ``Template`` instance or absolute coordinates (x, y)
    :param times: how many touches to be performed
    :param match: bool whether match template on screen. If False mean click the position about template.
    :param kwargs: platform specific `kwargs`, please refer to corresponding docs
    :return: finial position to be clicked, e.g. (100, 100)
    :platforms: Android, Windows, iOS
    :Example:
        Click absolute coordinates::

        >>> touch((100, 100))

        Click the center of the picture(Template object)::

        >>> touch(Template(r"tpl1606730579419.png", target_pos=5))

        Click 2 times::

        >>> touch((100, 100), times=2)

        Under Android and Windows platforms, you can set the click duration::

        >>> touch((100, 100), duration=2)

        Right click(Windows)::

        >>> touch((100, 100), right_click=True)

    """
    # assert fn_return is not None, "This function must have fn_return!"
    if isinstance(v, Template):
        if not match and v.record_pos:
            pos = Predictor.get_predict_point(v.record_pos, GetDevice().get_current_resolution())
        else:
            pos = ap_wait(v, timeout=timeout, threshold=threshold, interval=interval, snapshot=snapshot)
    else:
        try_log_screen()
        pos = v

    if not pos :
        return False

    for _ in range(times):
        GetDevice().touch(pos, **kwargs)
        time.sleep(0.05)
    time.sleep(delay if delay is not None else ST.OPDELAY)
    return pos

def ap_swipe(v1, v2=None, *, vector=None, match=True, duration=1.0, delay=None, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True, **kwargs):
    """
    这段代码修改自airtest.swipe，目的是使其支持根据图像坐标的模式，而不是必须要进行模板匹配

    Perform the swipe action on the device screen.

    There are two ways of assigning the parameters
        * ``swipe(v1, v2=Template(...))``   # swipe from v1 to v2
        * ``swipe(v1, vector=(x, y))``      # swipe starts at v1 and moves along the vector.


    :param v1: the start point of swipe,
               either a Template instance or absolute coordinates (x, y)
    :param v2: the end point of swipe,
               either a Template instance or absolute coordinates (x, y)
    :param vector: a vector coordinates of swipe action, either absolute coordinates (x, y) or percentage of
                   screen e.g.(0.5, 0.5)
    :param match: bool whether match template on screen.
    :param **kwargs: platform specific `kwargs`, please refer to corresponding docs
    :raise Exception: general exception when not enough parameters to perform swap action have been provided
    :return: Origin position and target position
    :platforms: Android, Windows, iOS
    :Example:

        >>> swipe(Template(r"tpl1606814865574.png"), vector=[-0.0316, -0.3311])
        >>> swipe((100, 100), (200, 200))

        Custom swiping duration and number of steps(Android and iOS)::

        >>> # swiping lasts for 1 second, divided into 6 steps
        >>> swipe((100, 100), (200, 200), duration=1, steps=6)

    """
    # assert fn_return is not None, "This function must have fn_return!"
    if isinstance(v1, Template):
        try:
            if not match and v1.record_pos:
                pos1 = Predictor.get_predict_point(v1.record_pos, GetDevice().get_current_resolution())
            else:
                pos1 = ap_wait(v1, timeout=timeout, threshold=threshold, interval=interval, snapshot=snapshot)
        except TargetNotFoundError:
            # 如果由图1滑向图2，图1找不到，会导致图2的文件路径未被初始化，可能在报告中不能正确显示
            if v2 and isinstance(v2, Template):
                v2.filepath
            raise
    else:
        try_log_screen()
        pos1 = v1

    if v2:
        if isinstance(v2, Template):
            if not match and v2.record_pos:
                pos2 = Predictor.get_predict_point(v2.record_pos, GetDevice().get_current_resolution())
            else:
                pos2 = ap_wait(v2, timeout=timeout, threshold=threshold, interval=interval, snapshot=snapshot)
        else:
            pos2 = v2
    elif vector:
        if vector[0] <= 1 and vector[1] <= 1:
            w, h = GetDevice().get_current_resolution()
            vector = (int(vector[0] * w), int(vector[1] * h))
        pos2 = (pos1[0] + vector[0], pos1[1] + vector[1])
    else:
        raise Exception("no enough params for swipe")

    GetDevice().swipe(pos1, pos2, duration=duration, **kwargs)
    time.sleep(delay if delay is not None else ST.OPDELAY)
    return pos1, pos2

ap_click = ap_touch

def ap_sleep(duration=1.0):
    """
    sleep 函数
    :param duration:
    :return:
    """
    time.sleep(duration)
    return True

def ap_text(_text, *, enter=True, delay=None, **kwargs):
    """
    这段代码修改自airtest.text

    Input text on the target device. Text input widget must be active first.

    :param text: text to input, unicode is supported
    :param enter: input `Enter` keyevent after text input, default is True
    :return: None
    :platforms: Android, Windows, iOS
    :Example:

        >>> text("test")
        >>> text("test", enter=False)

        On Android, sometimes you need to click the search button after typing::

        >>> text("test", search=True)

        .. seealso::

            Module :py:mod:`airtest.core.android.ime.YosemiteIme.code`

            If you want to enter other keys on the keyboard, you can use the interface::

                >>> text("test")
                >>> device().yosemite_ime.code("3")  # 3 = IME_ACTION_SEARCH

            Ref: `Editor Action Code <http://developer.android.com/reference/android/view/inputmethod/EditorInfo.html>`_

    """
    GetDevice().text(_text, enter=enter, **kwargs)
    time.sleep(delay if delay is not None else ST.OPDELAY)
    return True

def ap_keyevent(keyname, delay=None, **kwargs):
    """
    这段代码修改自airtest.keyevent
    Perform key event on the device

    :param keyname: platform specific key name
    :param **kwargs: platform specific `kwargs`, please refer to corresponding docs
    :return: None
    :platforms: Android, Windows, iOS
    :Example:

        * ``Android``: it is equivalent to executing ``adb shell input keyevent KEYNAME`` ::

        >>> keyevent("HOME")
        >>> # The constant corresponding to the home key is 3
        >>> keyevent("3")  # same as keyevent("HOME")
        >>> keyevent("BACK")
        >>> keyevent("KEYCODE_DEL")

        .. seealso::

           Module :py:mod:`airtest.core.android.adb.ADB.keyevent`
              Equivalent to calling the ``android.adb.keyevent()``

           `Android Keyevent <https://developer.android.com/reference/android/view/KeyEvent#constants_1>`_
              Documentation for more ``Android.KeyEvent``

        * ``Windows``: Use ``pywinauto.keyboard`` module for key input::

        >>> keyevent("{DEL}")
        >>> keyevent("%{F4}")  # close an active window with Alt+F4

        .. seealso::

            Module :py:mod:`airtest.core.win.win.Windows.keyevent`

            `pywinauto.keyboard <https://pywinauto.readthedocs.io/en/latest/code/pywinauto.keyboard.html>`_
                Documentation for ``pywinauto.keyboard``

        * ``iOS``: Only supports home/volumeUp/volumeDown::

        >>> keyevent("HOME")
        >>> keyevent("volumeUp")

    """
    GetDevice().keyevent(keyname, **kwargs)
    time.sleep(delay if delay is not None else ST.OPDELAY)
    return True

def ap_pinch(in_or_out='in', center=None, percent=0.5, delay=None):
    """
    这段代码修改自airtest.pinch
    Perform the pinch action on the device screen

    :param in_or_out: pinch in or pinch out, enum in ["in", "out"]
    :param center: center of pinch action, default as None which is the center of the screen
    :param percent: percentage of the screen of pinch action, default is 0.5
    :return: None
    :platforms: Android
    :Example:

        Pinch in the center of the screen with two fingers::

        >>> pinch()

        Take (100,100) as the center and slide out with two fingers::

        >>> pinch('out', center=(100, 100))
    """
    try_log_screen()
    GetDevice().pinch(in_or_out=in_or_out, center=center, percent=percent)
    time.sleep(delay if delay is not None else ST.OPDELAY)
    return True

def ap_keyboard(*key_code, internal=0.1, delay=None):
    """
    模拟键盘操作
    :param key_code: str
        "\t",
        "\n",
        "\r",
        " ",
        "!",
        '"',
        "#",
        "$",
        "%",
        "&",
        "'",
        "(",
        ")",
        "*",
        "+",
        ",",
        "-",
        ".",
        "/",
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        ":",
        ";",
        "<",
        "=",
        ">",
        "?",
        "@",
        "[",
        "\\",
        "]",
        "^",
        "_",
        "`",
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
        "{",
        "|",
        "}",
        "~",
        "accept",
        "add",
        "alt",
        "altleft",
        "altright",
        "apps",
        "backspace",
        "browserback",
        "browserfavorites",
        "browserforward",
        "browserhome",
        "browserrefresh",
        "browsersearch",
        "browserstop",
        "capslock",
        "clear",
        "convert",
        "ctrl",
        "ctrlleft",
        "ctrlright",
        "decimal",
        "del",
        "delete",
        "divide",
        "down",
        "end",
        "enter",
        "esc",
        "escape",
        "execute",
        "f1",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
        "f19",
        "f2",
        "f20",
        "f21",
        "f22",
        "f23",
        "f24",
        "f3",
        "f4",
        "f5",
        "f6",
        "f7",
        "f8",
        "f9",
        "final",
        "fn",
        "hanguel",
        "hangul",
        "hanja",
        "help",
        "home",
        "insert",
        "junja",
        "kana",
        "kanji",
        "launchapp1",
        "launchapp2",
        "launchmail",
        "launchmediaselect",
        "left",
        "modechange",
        "multiply",
        "nexttrack",
        "nonconvert",
        "num0",
        "num1",
        "num2",
        "num3",
        "num4",
        "num5",
        "num6",
        "num7",
        "num8",
        "num9",
        "numlock",
        "pagedown",
        "pageup",
        "pause",
        "pgdn",
        "pgup",
        "playpause",
        "prevtrack",
        "print",
        "printscreen",
        "prntscrn",
        "prtsc",
        "prtscr",
        "return",
        "right",
        "scrolllock",
        "select",
        "separator",
        "shift",
        "shiftleft",
        "shiftright",
        "sleep",
        "space",
        "stop",
        "subtract",
        "tab",
        "up",
        "volumedown",
        "volumemute",
        "volumeup",
        "win",
        "winleft",
        "winright",
        "yen",
        "command",
        "option",
        "optionleft",
        "optionright",
    :return:
    """
    for each in key_code:
        pyautogui.keyDown(each)
        time.sleep(0.05)
        pyautogui.keyUp(each)
        time.sleep(internal)
    time.sleep(delay if delay is not None else ST.OPDELAY)
    return True


# --- Overwrite ---
touch = ap_touch
click = ap_click
wait = ap_wait
exists = ap_exists
pinch = ap_pinch
swipe = ap_swipe
sleep = ap_sleep
keyevent = ap_keyevent
keyboard = ap_keyboard


if __name__ == '__main__':
    a = ...
    print(a is ...)



# !/usr/bin/env python
# -*- coding:utf-8 -*-
# @author :yinzhengjie
# blog:https://www.cnblogs.com/yinzhengjie

from setuptools import setup, find_packages, Extension

setup(
    # 指定项目名称，我们在后期打包时，这就是打包的包名称，当然打包时的名称可能还会包含下面的版本号哟~
    name='airpage',
    # 指定版本号
    version='0.1.3',
    # 这是对当前项目的一个描述
    description='based on airtest',
    # 作者是谁，指的是此项目开发的人，这里就写你自己的名字即可
    author='eaglebaby',
    # 作者的邮箱
    author_email='2229066748@qq.com',
    # 写上项目的地址，比如你开源的地址开源写博客地址，也开源写GitHub地址，自定义的官网地址等等。
    url='https://gitee.com/eagle-s_baby/airpage',
    # 指定包名，即你需要打包的包名称，要实际在你本地存在哟，它会将指定包名下的所有"*.py"文件进行打包哟，但不会递归去拷贝所有的子包内容。
    # 综上所述，我们如果想要把一个包的所有"*.py"文件进行打包，应该在packages列表写下所有包的层级关系哟~这样就开源将指定包路径的所有".py"文件进行打包!
    platforms='Windows',
    packages=find_packages(),
    python_requires='==3.9.*',
    # 需要安装的依赖
    install_requires=[
        'py_trees',
        'airtest',
        'pyautogui',
        'files3',
        'pywin32',
        "pydotplus",
        'wxauto',  #
    ],

    # 添加这个选项，在windows下Python目录的scripts下生成exe文件
    # 注意：模块与函数之间是冒号:
    entry_points={'console_scripts': [
        'ap_new = airpage.cmd_entry:CMD_NewProject',
        'ap_version = airpage.cmd_entry:CMD_Version',
    ]},

    # long_description=read('README.md'),
    classifiers=[  # 程序的所属分类列表
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License (GPL)",
    ],
    # 此项需要，否则卸载时报windows error
    zip_safe=False
)


from airpage.core.core import *

class Page:
    """
    定义了一个关于页面的节点模型

    * 有一个页面判别方法(Function)
    * 维护一个跳转字典，目标id:跳转方法(Action)
    * BFS页面跳转寻路

    --------------------------------------
    :param name: str 页面唯一标识符
    :param identity: Identity|list|tuple|Not|tpl 页面的逻辑判别，airpage定义的逻辑组合
            详细定义参见Core文件内的Identity对象
    """

    def __new__(cls, name, identity):
        # 检查name的重复性 & 将该Name注入公共空间
        env = GetEnviron()
        if env.GetPage(name) is None:
            tmp = super().__new__(cls)
            env.SetPage(name, tmp)
            return tmp
        else:
            error("DupublicNameError", f"You want to new a Page with dupublic name '{name}'.")

    def __init__(self, name, identity):
        self.name = name  # 一个Page的标识符
        self.identity: Identity = identity if isinstance(identity, Identity) else Identity(identity)  # 页面标识
        self.links = {}  # 页面跳转链接集 page:function
        self.tasks = {}  # 任务集 name: function

    def __str__(self):
        return f"Page:{self.name}"

    def __repr__(self):
        return f"Page:{self.name}"

    def check(self, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True):
        """
        检查当前画面是否是目标界面
        :return: bool
        """
        return self.identity.check(timeout=timeout, threshold=threshold, interval=interval, snapshot=snapshot)  # 只检查一次

    def _burstCheck(self, timeout=ST.FIND_TIMEOUT, interval=0.5, snapshot=True):
        """
        爆发式检测
        * 在短时间内没有任何!'操作'!
        * 在短时间内需要大量检测
        * 需要外部库内调用check方法(传参困难)
        :return:
        """
        return self.identity.check(timeout=timeout, interval=interval, snapshot=snapshot)

    def next(self, _next):
        """
        移动到相邻的页面
        :param _next: str|PageNode
        :return: bool
        """
        _next = _next if isinstance(_next, Page) else GetEnviron().GetPage(_next)

        _function = self.links.get(_next)

        if _function is None:
            error("MissingTransFunctionError", f"Miss transfunction for {self.name} -> {_next.name}")
            return False
        info(f"move Page:{self.name} next to Page:{_next.name}")
        return _function()

    def do(self, task_name, *args, **kwargs):
        """
        执行已经注册到当前PageNode中的任务
        :param task_name: 任务名称，与login时传入的名称相一致
        :param *args: 任务所需的参数
        :param **kwargs: 任务所需的kv参数
        :return: task的返回值 or None(找不到任务)
        """
        fn = self.tasks.get(task_name)
        if fn is not None:
            tmp = self.check(timeout=ST.FIND_TIMEOUT_TMP)
            # debug(f"Check:{tmp} when execute Task:{task_name} on Page:{self.name}")
            if not tmp:  # check and not self.check(timeout=ST.FIND_TIMEOUT_TMP):
                if not self.Transit(self.name):
                    return False
            info(f"Execute Task:{task_name} on Page:{self.name}")

            ''' --------- Specific --------- '''
            if isinstance(fn, Behaviour):
                kwargs['oneshot'] = True

            result = fn(*args, **kwargs)

            print(f"Execute Task:{task_name}  on Page:{self.name} Result:{result}")
            return result
        else:
            error("UnknownTask", f"Can not find Task:{task_name} in {self}")

    # 下面的段落描述了: 如何创建链接 & BFS链接搜索
    def link(self, page, action):
        """
        链接两个页面
        :param page: PageNode 目标页面
        :param action: 转移操作
        :return: None
        """
        print(f"Link Page:{self.name} to Page:{page.name}")
        assert isinstance(page, Page), "When you use 'link', the param1 must be a Page."
        if action is None:
            return
        if self.links.get(page):
            warn(f"You are replacing exist link:{page.name}")
        if page is self:
            error("LinkSelfError", "You can not link self to self.")
        self.links[page] = action

    def unlink(self, page):
        """
        断开与特定页面的链接
        :param page: PageNode 目标页面
        :return: None
        """
        self.links.pop(page)

    def login(self, name, function, check=True):
        """
        注册任务到指定页面
        :param name:
        :param action:
        :param check: 执行任务前是否需要检测当前screen是否为当前页面. 在dotask时你仍然有机会临时修改是否需要check
        :return:
        """
        if self.tasks.get(name):
            warn(f"You are replacing exist task:{name}")
        self.tasks[name] = function
        self.task_checks[name] = check

    def logoff(self, name):
        self.tasks.pop(name)
        self.task_checks.pop(name)

    @classmethod
    def Check(cls, page, *, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True):
        if isinstance(page, str):
            page = GetEnviron().GetPage(page)
        return page.check(
            timeout=timeout,
            threshold=threshold,
            interval=interval,
            snapshot=snapshot,
        )


    @staticmethod
    def ParallelCheck(sources=None, timeout=ST.FIND_TIMEOUT, interval=0.5):
        """
        根据当前界面推测对应的Page
        :param sources: list 所有参与推测的page
        :param timeout:
        :param interval:
        :return: page or None
        """

        info("Auto analyse page, It will take few time...")

        if sources is None:
            sources = list(GetEnviron()._pages.values())

        # 使用统一截图, 并创建pool
        sid = GetTempName()
        GlobalScreen_Update(sid)
        pool = ThreadPoolExecutor(max_workers=min(cpu_count * 2, len(sources)))

        # --- 创建check任务并提交到事件内核 ---
        tasks = [
            pool.submit(page._burstCheck, ST.FIND_TIMEOUT_TMP, interval, sid) for page in sources
        ]

        # --- 等待任务结束 ---
        _, results = wait_results(tasks, timeout, interval)

        # 清空全局截图，同时释放pool
        GlobalScreen_Flush(sid)
        pool.shutdown()

        # --- 读取各个Page.check返回的结果 ---
        for i, item in enumerate(results):
            if item:
                info(f"It's probably on Page:{sources[i]}")
                return sources[i]

        warn("Can not find a page match with screen now.")
        return

    ''' ------------------------------------------------- '''


    @staticmethod
    def _bfs_path_found(page_a, page_b):
        """
        寻找由page_a到page_b的路径
        :param page_a:
        :param page_b:
        :return: list of page, head is page_a, end is page_b.
                * return [] if failed.
        """

        def _inner_(from_node, target, excepts):
            """
            实现一轮迭代，如果找到结果，返回target; 重复输入返回[]; 其余情况返回下一轮的queue:[]
            :param from_node:
            :param target:
            :param excepts:
            :return:
            """
            if target in excepts:
                return []

            excepts.append(from_node)

            queue = []
            for each in from_node.links:
                if each is target:
                    return target
                queue.append(each)

            return queue

        g_queue, excepts = [([page_a], page_a), ], []
        while True:
            _ = []
            for path, each in g_queue:
                # print(path, each, page_b, excepts)
                queue = _inner_(each, page_b, excepts)

                if isinstance(queue, Page):  # 函数出口
                    return path + [page_b]

                for each in queue:  # 加入下一次循环队列
                    _.append((path + [each], each))  # 基本单元是(path, next_node)

            if not _:  # 未能寻找到结果
                return []

            g_queue = _

    ''' ---------------------------------------------------------- '''

    @staticmethod
    def LinkWith(page_name_a, page_name_b, action_a2b=None, action_b2a=None):
        """
        创建一个界面与另一个界面的链接
        :param page_a:
        :param page_b:
        :param action_a2b: None表示无法跳转
        :param action_b2a: None表示无法回退
        :return:
        """
        page_a = GetEnviron().GetPage(page_name_a)
        page_b = GetEnviron().GetPage(page_name_b)

        page_a.link(page_b, action_a2b)
        page_b.link(page_a, action_b2a)

    @staticmethod
    def DoTask(page_name, task_name, *args, check=None, **kwargs):
        """
        执行已经注册到当前Page中的任务
        :param page_name:
        :param task_name: 任务名称，与login时传入的名称相一致
        :param *args: 任务所需的参数
        :param check: 执行任务前是否需要检测当前screen是否为当前页面
        :param **kwargs: 任务所需的kv参数
        :return: task的返回值 or None(找不到任务)
        """
        return GetEnviron().GetPage(page_name).do(task_name, *args, check=check, **kwargs)

    @staticmethod
    def LoginTask(page_name, fn):
        """
        注册任务到指定页面.
        :param page_name:
        :param fn:
        :return:
        """
        return GetEnviron().GetPage(page_name).login(fn.__name__, fn)

    @classmethod
    def Transit(cls, arg1, arg2=None, timeout=ST.FIND_TIMEOUT, interval=0.5):
        """
        移动当前页面到指定页面
        * arg2 is None时，arg1表示目标页面，会自动检查当前页面
        * arg1和arg2均存在时，表示开发人员已确定当前页面为arg1，arg2为目标页面
        :param arg1: str|PageNode
        :param arg2: str|PageNode
        :return: bool
        """
        env = GetEnviron()

        # 检查参数的个数，判断用户的意图(1个参数: 参数表示终点，起点需要程序寻找; 2个参数，起点和终点均已给出)
        arg1 = arg1 if isinstance(arg1, Page) else env.GetPage(arg1)
        arg2 = None if arg2 is None else (arg2 if isinstance(arg2, Page) else env.GetPage(arg1))

        # 检查参数的类型, 确定当前页面(如果需要搜索当前页面的话)，并将其统一转为Page对象
        page_a = cls.ParallelCheck(timeout=timeout, interval=interval) if arg2 is None else arg1
        page_b = arg1 if arg2 is None else arg2

        # 检查第一个参数(如果参数二存在时)
        if arg2 is not None:
            if not page_a.check():
                warn(f"检查页面起点:{page_a}失败. 由页面:{page_a}到{page_b}的操作未能完成.")
                return False

        if page_a is None:
            return False

        if page_a.name == page_b.name:
            return True

        _path = Page._bfs_path_found(page_a, page_b)
        end_page = _path[-1]

        if not _path:
            warn(f"搜寻路径{page_a}->{page_b}失败. 由页面:{page_a}到{page_b}的操作未能完成.")
            return False

        if len(_path) < 2:
            error("UnexpectedShortLength", "len of path can not shorter than 2")

        _flag = True
        for i in range(len(_path[:-1])):
            _flag = _flag and _path[i].next(_path[i + 1])

            if not _flag:
                break

            sleep(ST.FIND_TIMEOUT_TMP)

        if not _flag or not end_page.check(timeout, interval):
            warn(f"执行页面:{_path[i]}->{_path[i + 1]}的过渡函数失败. 由页面:{page_a}到{page_b}的操作未能完成.")
            return False

        return True

    @staticmethod
    def RenderTask(page_name, task_name, *render_args, **render_kwargs):
        return GetEnviron().GetPage(page_name).tasks[task_name].render(*render_args, name=f"{page_name}_{task_name}",
                                                                        **render_kwargs)

    def __call__(self, fn: typing.Callable):
        """
        装饰任务函数。被装饰的函数会被当作一个Task
        :param fn: 任务函数
        :param args: fn参数
        :param kwargs: fn的kv参数
        :return:
        """
        self.login(fn.__name__, fn)

        def _inner_(*args, **kwargs):
            return self.do(fn.__name__, *args, **kwargs)

        return _inner_

''' --------------------------------------------------------------------------------------------------------------- '''


# Task装饰器
def page(page_name: str) -> Page:
    """
    获得特定名称的page
    * Page的实例其实是对Task的装饰器
    :param page_name: page的名称
    :return: Page - wrapper
    """
    return GetEnviron().GetPage(page_name)


# Link装饰器
def transfer(page_a: typing.Union[str, Page], page_b: typing.Union[str, Page]) -> typing.Callable:
    """
    装饰从page_a转移到page_b的函数
    :param page_a: page a或其名称
    :param page_a: page b或其名称
    :return : wrapper fn
    """

    _page_a = page_a
    _page_b = page_b

    def _wrapper_(fn):
        page_a = _page_a if isinstance(_page_a, Page) else GetEnviron().GetPage(_page_a)
        page_b = _page_b if isinstance(_page_b, Page) else GetEnviron().GetPage(_page_b)
        page_a.link(page_b, fn)
        return fn

    return _wrapper_



import airpage.core._btree as bt

from math import inf
from airpage.core.page import *

import pydotplus as pydot

def ResetMousePosition():
    pyautogui.moveTo(1, 1)

def ResetMousePositionAndFlush():
    pyautogui.moveTo(1, 1)
    GlobalScreen_Flush()

def screen_flush():
    return action(
        GlobalScreen_Flush,
        _reback=3,  # pack时需要额外往后移动一级
    )

def screen_update():
    return action(
        GlobalScreen_Update,
        _reback=3,  # pack时需要额外往后移动一级
    )

def condition_wait(querys, *, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True):
    return condition(
        wait,
        querys,
        timeout=timeout,
        threshold=threshold,
        interval=interval,
        snapshot=snapshot,
        _reback=3, # pack时需要额外往后移动一级
    )

def condition_exists(querys, *, threshold=None, snapshot=True):
    return condition(
        exists,
        querys,
        threshold=threshold,
        snapshot=snapshot,
        _reback=3, # pack时需要额外往后移动一级
    )
# 默认标准的检查时间
def action_touch(v, *, times=1, match=True, delay=None, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True, **kwargs):
    return action(
        touch,
        v,
        times=times,
        match=match,
        delay=delay,
        timeout=timeout,
        threshold=threshold,
        interval=interval,
        snapshot=snapshot,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
        ** kwargs
    )

# 默认较短的检查时间
def action_tickle(v, *, times=1, match=True, delay=None, timeout=ST.FIND_TIMEOUT_TMP, threshold=None, interval=0.5, snapshot=True, **kwargs):
    return action(
        touch,
        v,
        times=times,
        match=match,
        delay=delay,
        timeout=timeout,
        threshold=threshold,
        interval=interval,
        snapshot=snapshot,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
        ** kwargs
    )

def action_sleep(duration=1.0):
    return action(
        ap_sleep,
        duration,
        _reback=3, # pack时需要额外往后移动一级
    )

def action_swipe(v1, v2=None, *, vector=None, match=True, duration=1.0, delay=None, timeout=ST.FIND_TIMEOUT_TMP, threshold=None, interval=0.5, snapshot=True, **kwargs):
    return action(
        swipe,
        v1,
        v2,
        vector=vector,
        match=match,
        duration=duration,
        delay=delay,
        timeout=timeout,
        threshold=threshold,
        interval=interval,
        snapshot=snapshot,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
        **kwargs
    )


def action_pinch(in_or_out='in', center=None, percent=0.5, delay=None):
    return action(
        pinch,
        in_or_out=in_or_out,
        center=center,
        percent=percent,
        delay=delay,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
    )


def action_text(_text, *, enter=True, delay=None, **kwargs):
    return action(
        text,
        _text,
        enter=enter,
        delay=delay,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
        **kwargs
    )


def action_keyevent(keyname, *, delay=None, **kwargs):
    return action(
        keyevent,
        keyname,
        delay=delay,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
        **kwargs
    )

def action_keyboard(*key_code, internal=0.1, delay=None):
    return action(
        keyboard,
        *key_code,
        internal=internal,
        delay=delay,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
    )

action_click = action_touch


''' --------------------------AP Action End------------------------------ '''

def action_label(desc):
    return action(label, desc, _reback=3)

def action_info(*args):
    return action(info, *args, _reback=3)

def action_warn(*args):
    return action(warn, *args, _reback=3)

def action_error(err_type:str, *args):
    return action(error, err_type, *args, _reback=3)

''' --------------------------AP Information End------------------------------ '''

def condition_identity(identity, *, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True):
    return condition(
        Identity.Check,
        identity,
        timeout=timeout,
        threshold=threshold,
        interval=interval,
        snapshot=snapshot,
        _reback=3, # pack时需要额外往后移动一级
    )


def condition_page(page, *, timeout=ST.FIND_TIMEOUT, threshold=None, interval=0.5, snapshot=True):
    return condition(
        Page.Check,
        page,
        timeout=timeout,
        threshold=threshold,
        interval=interval,
        snapshot=snapshot,
        _reback=3,  # pack时需要额外往后移动一级
    )

def action_transit(*args):
    return action(
        Page.Transit, *args,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
    )

def action_dotask(page_name, task_name, *args, **kwargs):
    return action(Page.DoTask, page_name, task_name,
        *args,
        **kwargs,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
    )

''' --------------------------AP Composite Action&Condition End------------------------------ '''

def action_fwrite(file, string):
    """
    文件写入操作
    * 可以是一般文件
    * 可以是标准输出
    * 可以是自定义输出
    :param file: 具有方法: .write(__s)的任意对象
    :param string: 待写入的字符串
    :return: Behaviour
    """
    return action(
        file.write,
        string,
        delay=delay,
        _sys_after_ = ResetMousePositionAndFlush,  # 这类action可能会改变屏幕，所以清除掉屏幕，强迫更新屏幕
        _reback=3, # pack时需要额外往后移动一级
    )

def AP_Polling(btree:bt.Behaviour, oneshot=False):
    """
    轮询行为树, 并返回轮询结果
    * 具有AP特色
    :param btree: BehaviourTree
    :return: bool or ...
    """
    while True:
        btree.tick_once()
        result = btree.status

        if result != bt.Status.RUNNING or not oneshot:
            break

    return FromStatus(result)

bt.SetTreeEntryence(AP_Polling)

''' --------------------------AP Polling End------------------------------ '''

_default_values=dict(
    v2=None,
    vector=None,
    times=1,
    match=True,
    delay=None,
    timeout=ST.FIND_TIMEOUT,
    threshold=None,
    interval=0.5,
    snapshot=True,
    duration=1.0,
)

def CreateNode(graph: pydot.Dot, behaviour: bt.Behaviour, root=False):
    def create_template_subnodes(_graph, _node_name, *templates):
        """
        创建有关Template的节点，传入Template以及可能包含Template的列表
        * 返回是否实际处理了Template
        :return: bool
        """
        _result = False
        for tpl in templates:
            if isinstance(tpl, Template):
                _i = bt.GetNewId('Template')
                graph.add_node(pydot.Node(
                    name=f"{_node_name}_tpl{_i}",
                    shape="plaintext",
                    label='',
                    image=os.path.join(ProjectPath(), tpl.filename),
                ))

                graph.add_edge(pydot.Edge(
                    name, f"{name}_tpl{_i}",
                    style='dashed',
                    arrowhead='none',
                ))
                _result |= True
            elif isinstance(tpl, (list, tuple)):
                _result |= create_template_subnodes(_graph, _node_name, *tpl)
        return _result

    def is_pos(item):
        if not isinstance(item, (list, tuple)):
            return False

        _flag = True
        for each in item:
            if not isinstance(each, (int, float)):
                _flag = False
                break

        return _flag

    name = f"node{bt.GetNewId('_node')}"
    if not root and hasattr(behaviour, 'packer'):
        node = pydot.Node(
            name=name,
            label=behaviour.packer,
            shape="folder",
            style="filled",
            fillcolor="khaki",
            fontsize=16,
            fontcolor="black",
        )

        graph.add_node(node)
    elif isinstance(behaviour, (bt.Action, bt.Condition)):
        # 确定Action和Condition节点的名称，并对ap_xxx的名称的显示进行优化
        _label = behaviour.target.__name__ if behaviour.target else behaviour.name
        if len(_label) >= 3 and _label[0:3] == 'ap_':
            _label = _label[3:]

        node = pydot.Node(
            name=name,
            label=_label,
            shape="ellipse" if isinstance(behaviour, bt.Condition) else "box",
            style="filled",
            fillcolor="lightgray",
            fontsize=12,
            fontcolor="black",
        )
        graph.add_node(node)

        ''' --- 为action和condition创建参数节点 --- '''
        args, kwargs = behaviour.target_args, behaviour.target_kwargs
        any_flag = False  # 用于判断是否需要渲染参数节点

        # 生成label
        # 检查有无tpl
        _label, _i = "", 0
        for a in args:
            if isinstance(a, Template) or isinstance(a, (list, tuple)):
                if is_pos(a):
                    any_flag |= True
                    _label += f"{a}, \n"
                else:
                    create_template_subnodes(graph, name, a)
            else:
                any_flag |= True
                if isinstance(a, str):
                    a = f"'{a}'"
                _label += f"{a}, \n"

        for k, v in kwargs.items():
            if k not in _default_values or _default_values[k] != v:
                if isinstance(v, Template) or isinstance(v, (list, tuple)):
                    if is_pos(v):
                        any_flag |= True
                        _label += f"{k}={v}, \n"
                    else:
                        create_template_subnodes(graph, name, v)
                else:
                    any_flag |= True
                    if isinstance(v, str):
                        v = f"'{v}'"
                    _label += f"{k}={v}, \n"

        # 只有实际处理了参数的节点才进行显示
        if any_flag:
            if len(_label) >= 3 and _label[-3:] == ', \n':
                _label = _label[:-3]

            graph.add_node(pydot.Node(
                name=f"{name}_param",
                label=_label,
                shape="ellipse",
                style="filled",
                fillcolor="ghostwhite",
                fontsize=10,
            ))

            graph.add_edge(pydot.Edge(
                name, f"{name}_param",
                style='dashed',
                arrowhead='none',
                fontsize=8,
            ))

    elif isinstance(behaviour, bt.Composite):
        isi_seq = isinstance(behaviour, bt.Sequence)
        isi_par = isinstance(behaviour, bt.Parallel)
        # isi_fab = isinstance(behaviour, bt.Selector)  # No need to if it

        node = pydot.Node(
            name=name,
            label=' -> ' if isi_seq else (" => " if isi_par else " ? "),
            shape="square",
            style="filled",
            fillcolor='orange' if isi_seq else ("gold" if isi_par else "tomato"),
            fontsize=16,
            fontcolor="black",
        )

        graph.add_node(node)

        for child in behaviour.children:
            subnode = CreateNode(graph, child)
            graph.add_edge(pydot.Edge(
                name, subnode.obj_dict['name']
            ))

    elif isinstance(behaviour, bt.Decorator):

        _label = behaviour.name

        if isinstance(behaviour, bt.Repeater):
            _label = "While\n"
            _tmp = ""
            if behaviour.max_times is not inf:
                _tmp += f"{behaviour.max_times}, "
            if behaviour.until is not None:
                _tmp += f"{bt.FromStatus(behaviour.until)}"
            if _tmp:
                _label += f"({_tmp})"
        elif isinstance(behaviour, bt.Reflictor):
            _label = "Fn\n"
            _label += f"({bt.FromStatus(behaviour.status_x)}->{bt.FromStatus(behaviour.status_y)})"
        elif isinstance(behaviour, bt.Timeout):
            _label = "Time\n"
            _label += f"({behaviour.duration}s)"
        elif isinstance(behaviour, bt.EternalGuard):
            _label = "If\n"
            if behaviour.condition:
                _label += f"({behaviour.condition.__name__}s)"
        elif isinstance(behaviour, bt.Succeedor):
            _label = "True"
        elif isinstance(behaviour, bt.Inverter):
            _label = "Not"
        elif isinstance(behaviour, bt.OneShot):
            _label = "Oneshot"

        node = pydot.Node(
            name=name,
            label=_label,
            shape="diamond",
            style="filled",
            fillcolor='skyblue',
            fontsize=10,
            fontcolor="black",
        )

        graph.add_node(node)

        for child in behaviour.children:
            subnode = CreateNode(graph, child)
            graph.add_edge(pydot.Edge(
                name, subnode.obj_dict['name']
            ))
    else:
        node = pydot.Node(
            name=name,
            label=str(behaviour),
            shape="ellipse",
            style="filled",
            fillcolor="gray",
            fontsize=10,
            fontcolor="black",
        )
        graph.add_node(node)

    return node

def AP_Render(btree:bt.Behaviour, name=None, open=False):
    """
    渲染行为树
    * 具有AP特色
    :param btree: BehaviourTree
    :param name: 指定文件名，默认None，即使用节点名称
    :param open: bool 是否立刻打开svg文件(因为svg文件比png清楚)
    :return: None
    """
    graph = pydot.Dot(graph_type="digraph", ordering="out")
    CreateNode(graph, btree, root=True)
    output_path_noext = os.path.join(ProjectPath(), btree.name if name is None else name)
    graph.write_png(output_path_noext + '.png')
    graph.write_svg(output_path_noext + '.svg')

    if open:
        os.startfile(output_path_noext + '.svg')


bt.SetTreeRender(AP_Render)

''' --------------------------AP Render End------------------------------ '''


if __name__ == '__main__':
    def if_something(sth):
        return bool(sth)

    def do_something(sth):
        return True

    t = fallback(
    sequence(
    condition(if_something, "条件1"),
    repeater(action(do_something, "事请1"), times=3),
    ),
    action(do_something, "事请2"),
    )

    t.render(name="tree", open=True)
    print("ok")